
;;;;  compile-file
;;;;
(in-package "compiler")

(define *compile-name* "")  ;;; reverse.scm-> "reverse"
;;;(define *comp-out-file* "")

(define *comp-buff* (make-string 1000))

(define (load-form s)
  (let ((out  (open-output-string *comp-buff* ))(in 0)  )
    (write s out)
    (close-port out)
    (set! in (open-input-string *comp-buff*))
    (catch (load in))   ;;;;;; 04.5.17
    (close-port in)
 )    
)

(define (special-begin? x)
     (set! x (special? x))
    (if (eq? x 'begin)
        x 
        #f
    )
)

(define (comp-load file flag)
  (let ((s 0)(f 0)(out #f))
     (set! s (open-input-file (find-scm file )))
     (when flag
        (set! out (open-output-file (string-append *compile-name* ".sr")))
         (format  #t "out ~s~%" out)
        (write :BEGIN-SR out)
        (newline out)
     ) 
         (do ((x (read s)(read s)))
           ((eof-object? x)
                  (close-input-port  s)
                )
           (if (and (not flag) (pair? x))
                (begin
                   (set! f (special? (car x)))
                   (if (special-begin? (car x))
                      (compiler-eval-define-begin  x );; pass org
                      (compiler-eval-define x (macro-expand-rec x ()));; pass org&expand
                   )
                )
                (if flag 
                    (begin
         ;;;(format  #t "read x ~s~%" x)
                        (set! x (macro-expand-rec x ()))
         ;;;(format  #t "expan x ~s~%" x)
                        (compiler-eval-package  x)
                        (sr-out  (compile-process  x) out)
                    )
                )
             )
           )
   (when out
        ;;;(sr-out-sub '(:EXIT) out)
        (write :END-SR out)
        (close-output-port  out)
   )
))

(define (compiler-eval-define-begin x)
   (let* ((f 0))
    (set! x (cdr x))
    (while (pair? x)
        (set! f (car x))
        (if (special-begin? (car f))
            (compiler-eval-define-begin f) 
            (compiler-eval-define f  (macro-expand-rec x ())) ;;; toriaezu
        )
        (set! x (cdr x))
    )
))

(define (compiler-eval-define org x)   ;;; x is expanded of org
  (if (pair? x)
     (let* ( (f (car x))(sp 0))
         ;;;  f may be alias
         (set! sp (special? f))
         (if sp 
             (set! f sp)
         )
        (when (eq? 'define-syntax (car org))
              (set! f 'define-syntax)
                   ;;;; define-syntax is expanded (begin ...)
                              ;;; so we need car  org
         )
       (format #t "compiler-eval-define ~s ~s~%" f x) ;;;; for debug

        (cond ((eq? f 'begin)
                  (compiler-eval-define-begin x)
                )
              ((eq? f 'load)
                          (compiler-load-define (cadr x)))  ;;;;load file 04.5.3
                         ;;; (comp-load (cadr x) #f)
              ((eq? f 'compiler-eval) (load-form (cons 'begin (cdr x))))
              ((eq? f 'compiler-not-eval) )
              ((eq? f 'define) (load-form  x))
              ((eq? f 'define-macro) (load-form  x))
              ((eq? f 'define-syntax) (load-form  x))  ;;;;needs! 04.5.3
              ((eq? f 'defstruct) (load-form  x))  ;;;;needs! 04.5.3
              ((eq? f 'require) (load-form  x))
              ((eq? f 'provide) (load-form  x))
              ((eq? f 'remove-module) (load-form  x))
              (else
                 (compiler-eval-package x ) )
          )
      )
  )
)

(define (compiler-eval-package x)
  (if (pair? x)
     (let* ((f (car x)))
        (cond ((eq? f 'begin)
                    (set! x (cdr x))
                 (while (pair? x)
                    (compiler-eval-package (car x))
                    (set! x (cdr x))
                 )
                )
                         ;;; (comp-load (cadr x) #f)
              ((or  (eq? f 'in-package) (eq? f 'export)(eq? f 'use-package)
                    (eq? f 'intern)(eq? f 'shadowing-intern)(eq? f 'unintern)
                    (eq? f 'unuse-package))
                    (catch (eval x))
                  )
              (else )
          )
      )
  )
)
                     

(define (find-scm file)
  (let ((s 0))
  (set! file (find-source file #t))   ;;; s_load.c
  (if file (set! *compile-name* (car (file-sepa-ext file)))
        (error 1222 file)  )
   file
))


(define (compiler op . rest)
  (let ()
    (cond ((eq? op :load ) (compiler-load rest))
          ((eq? op :define) (map compile rest))
          (#t (comp-error 205 (list 'compiler op)  ))
    )
))

(define (compiler-load files)
  (while (pair? files)
    (comp-load (pop files) #f)
))

(define (compile-file file)
     (set! *verbose* #f)
     (comp-load file #f)
    ;;; (init-compile)
     ;;;(in-package "compiler")
     (set! *verbose* #t)
     (comp-load file #t)
     '***end-of-compile-file***
)

(define (sr-out comp out)
  (display #\( out)
  (while (pair? comp)
    (sr-out-sub (car comp) out)
    (set! comp (cdr comp))
  )
  (display #\) out)
  (newline out)
)
;;;; 
(define (sr-out-sub x out)
  (let ((op (car x))(s 0))
    (set! op 
       (cond ((eq? op :FUNC) :F)
             ((eq? op :STACK-NARG) :N) 
             ((eq? op :EXIT) :Q) 
             ((eq? op :JSR) (list :J (cadr x)))
             ((eq? op :TAIL-JUMP) (list :TJ (cadr x)))
             ((eq? op :JMP) (list :JM (cddr x)))
             ((eq? op :DEFINE) (list :D (cadr x)))
             ((eq? op :SET)
                (set! s (cadr x))
                (cond ((eq? s :FRAME) (list :S (car (caddr x))) )
                      ((eq? s :TOP)   (list :T (caddr x)))
                      ((eq? s :ENV) (list :V  (car (caddr x))) )
                      (else (error 5999 x))
                )
               )
              ((eq? op :LOAD)
                (set! s (cadr x))
                (cond ((eq? s :FRAME) (list :L (car (caddr x))) )
                      ((eq? s :TOP)   (list :P (caddr x)))
                      ((eq? s :ENV) (list :E  (car (caddr x))) )
                      ((eq? s :IMMIDIATE)   (list :I (caddr x)))
                      ((eq? s :REG) (cdr x) )  ;; :REG :ENV
                      ((pair? (cddr x))
                          (error 6999 x))
                      ((eq? s :UNDEF) :u)
                      ((eq? s :T) :t)
                      ((eq? s :F) :f)
                      ((eq? s ()) :n)
                      ((eq? s 0) :0)
                      ((eq? s 1) :1)
                      ((eq? s -1) :-1)
                      (else (error 7999 x))
                )
               )
             ((eq? op :IF)
                (set! s (cadr x))
                (if (eq? s :T)
                     (list :IT (cdddr x))
                     (list :IF (cdddr x))
                )
               )
             ((eq? op :BLOCK-IN) :K)
             ((eq? op :BLOCK-OUT) :O)
             ((eq? op :BIND) (list :B (caadr x)))
             ((eq? op :LETBIND) (list :LT (caadr x)))
             ((eq? op :LABEL) #f)  ;;; No out
             ((eq? op :CALL/CC) :C) 
             ((eq? op :LAMBDA-BEGIN) (list :LM (caddr x)))
             ((eq? op :LAMBDA-BIND) (list :LB (cadr x)(caddr x)))
             ((eq? op :ALLOC) (list :A (cadr x)))
             ((eq? op :DEALLOC) (list :DA (cadr x)))
             ((eq? op :MAKE-LAMBDA) (list :ML (cadr x)))
             ((eq? op :MAKE-MACRO) (list :MM (cadr x)))
             (else (error 8999 x))
        )
     )
     (when op
        (if (pair? op)
              (if (eq? (car op) :LB)
                  (format out " ~s ~s ~s" (car op) (cadr op)(caddr op))
                  (format out " ~s ~s" (car op) (cadr op))
              )
              (format out " ~s"  op)
         )
     )
))
              
;;;; for debug ,  mctest   mc.chk
(define m 0)

(define (mctest z)
  (set!  m (make-mc z (length z)))
  (mc-eval m ())
)

                 
