;;;  pretty-print
;;;  2003.11.20
;;;  2007.1.21 revised for infinity list

;;;(compiler-not-eval (load "write-loop-list"))
(load "write-loop-list")

(define *pp-level* 0)
(define *top-flag* #t)
(define *paren* #f)

(define (pp s . opt)
  (let* ((wport ()))
    (if (pair? opt)
        (set! wport  (car opt))
        (set! wport (car *stdout-stack*))
    )
  (set! *pp-level* 0)
  (format wport "~&")
  (set! *top-flag* #t)
  (pp-sub s wport #t)
))

(define (pp-sub s wp top-f)
   (if (pair? s)
         (begin
           (set! s (write-list-form s))
           (pp-form (car s) (cdr s) wp)
         )
       (pp-atom s wp top-f)
   )
)

(define (pp-atom s wp top-f)
   (let* ((pos (get-line-pos wp)))
     (if (> pos 65) 
       (begin
         (newline wp)
         (pp-indent wp)
       )
       (if (not top-f)
         (display #\space wp)
       )
    )
  (pp-atom-print s wp)
  (set! *top-flag* #f)
))

(define (pp-atom-print s wp)
  (cond ((vector? s) (pp-vector s wp))
        ((matrix? s) (pp-matrix s wp))
        ((string? s) (pp-string s wp))
        (else (write s wp))
  )
)

(define (pp-vector s wp)
  (let* ((n *pp-level*))
    (set! *pp-level* (+ n 1))
    (display #\# wp)
    (pp-sub (vector->list s) wp #f)
    (set! *pp-level*  n )
))

(define (pp-string s wp)
  (let* ((pos (get-line-pos wp)) (k 0) )
    (set! k(string-length s))
    (set! pos (+ pos k))
    (when  (> pos 75) 
       (newline wp)
       (pp-indent wp 1)
    )  
    (write s  wp)
))

(define (pp-matrix s wp)
  (let* ((n *pp-level*) (ma (matrix->list s)))
    (set! *pp-level* (+ n 1))
    (newline wp)
    (pp-indent wp -1)
    (display "#a(" wp)
    (while (pair? ma)
      (pp-sub (car ma) wp #t)
      (set! ma (cdr ma))
      (when (pair? ma)
        (newline wp)
        (pp-indent wp)
      )
    )
    (display #\) wp)
    (newline wp)
    (set! *pp-level*  n )
    (pp-indent wp)
))

(define (block-start  wp)
      (if *top-flag*
             (display "("  wp)
             (begin
                   (newline wp)
                   (pp-indent wp)
                   (display #\( wp)
             )
      )
  (set! *top-flag* #f)
)

(define (check-line pos lim wp)
    (when (> pos lim)                     
         (newline wp)
         (pp-indent wp)
       )
)
     
(define (pp-form top  rest  wp )
   (let* ((pos (get-line-pos wp)) (rightp #t)(block #t)(define-f #f))
     (set! *pp-level* (+ *pp-level* 1))
     (check-line pos 65 wp)
     (cond ((and (eq? top 'quote) (eq? (cdr rest) ()))
                (set! rightp #f)
                (display " '" wp) (pp-sub (car rest) wp #t)  )
           ((and (eq? top 'quasiquote) (eq? (cdr rest) ()))
                (set! rightp #f)
                (display " `"  wp) (pp-sub (car rest) wp #t)  )
           ((and (eq? top 'unquote) (eq? (cdr rest) ()))
                (set! rightp #f)
                (display " ,"  wp) (pp-sub (car rest) wp #t)  )
           ((and (eq? top 'unquote-splicing) (eq? (cdr rest) ()))
                (set! rightp #f)
                (display " ,@" wp) (pp-sub (car rest) wp #t)  )
           ((and (or (eq?  top 'define)(eq? top 'define-macro)) (pair? (car rest)))
               (set! define-f #t)
               (block-start  wp) 
               (format wp  "~s " top)
               (pp-sub  (car rest) wp #t)
               (newline wp)
               (set! rest (cdr rest))
               (while (pair?  rest) 
                  (pp-indent wp)
                  (pp-sub (car rest) wp #t)
                    (set! rest (cdr rest))
                 (if (pair?  rest)
                      (newline wp)
                 )
                )
            )
           ((or (eq? top 'if) (eq? top 'while) (eq? top 'do) (eq? top 'switch) (eq? top 'when)
                   (eq? top 'let)(eq? top 'let*)(eq? top 'letrec) (eq? top 'lambda))
              (pp-if top rest wp)
              )
           ((eq? top 'cond)
               (check-line pos 50 wp)
                (pp-cond top rest wp) )
           ( else
              (set! block #f)
               (check-line pos 50 wp)
              (display "(" wp)
              (pp-sub top wp #t)
               (while (pair?  rest)
                 (if (= (get-line-pos wp) 0) (pp-indent wp))
                 (pp-sub (car rest) wp #f)
                 (set! rest (cdr rest))
                )
               (when (not (eq? rest ()))  ;;;;Dot pair
                 (display " . " wp )
                 (pp-sub rest wp #f)
               )
              )
        )
     (set! *pp-level* (- *pp-level* 1))
     (when rightp
       (if (= (get-line-pos wp) 0) (pp-indent wp))
       (display ") " wp)
       (if (and define-f (= *pp-level* 0))
           (newline wp)  )
       (if block 
              (begin (newline wp) (set! *top-flag* #t))
              (set! *top-flag* #f)
       )
     )
))

(define (pp-if top rest wp)
       (block-start  wp)
               (format wp  "~s" top)
                (pp-sub (car rest) wp #f)
                (newline wp)
               (set! rest (cdr rest))
               (while (pair?  rest)
                 (pp-indent wp)
                 (pp-sub (car rest) wp #t)
                 (set! rest (cdr rest))
                 (if (and (pair?  rest) (not *top-flag*))
                      (newline wp)
                 )
                )
)

(define (pp-cond-clause  s wp)
  (let* ((c (car s)) (rest (cdr s)))
       ;;;(block-start  wp)
         (if *top-flag* (pp-indent wp))
               (display #\( wp)
               (pp-sub c wp #t)
           (set! *pp-level* (+ *pp-level* 1))
               (while (pair?  rest)
                 (newline wp)
                 (pp-indent wp)
                 (pp-sub (car rest) wp #t)
                 (set! rest (cdr rest))
                )
           (set! *pp-level* (- *pp-level* 1))
         (if *top-flag* (pp-indent wp))
         (display #\) wp)
         (newline wp)
         (set! *top-flag* #t)
))

(define (pp-cond top rest wp)
      (block-start  wp)   
      (pp-sub  top wp #t)
     (set! *pp-level* (+ *pp-level* 1))
           ;;;    (set! rest (cdr rest))
       (while (pair?  rest)
         ;;;   (pp-indent wp)
             ;;(pp-sub (car rest) wp #f)
             (pp-cond-clause  (car rest) wp )
             (set! rest (cdr rest))
             (when (and (pair?  rest) (not *top-flag*))
                   (newline wp)
                   (pp-indent wp)
              )
        )
     (set! *pp-level* (- *pp-level* 1))
)
          
(define (pp-indent wp . ad)
   (let* ((n *pp-level*))
     (if (pair? ad) (set! n (+ n (car ad))))
     (while (> n 0)
        (display "   " wp)
        (set! n (- n 1))
     )
  (set! *top-flag* #t)
))
         
;;;  file  output
(define (pp-file  in-file . out-f)
  (let* ((out 0)(in 0) (s 0))
    (if (pair? out-f)
       (set! out (open-output-file (car out-f)))
       (set! out (car *stdout-stack*))
    )
   (set! in (open-input-file in-file))
      (set! s (read in))
   (while (not (eof-object? s))
      (pp s out)
      (set! s (read in))
   )
   (if (pair? out-f) (close-port out))
   (close-port in)
))
           
     

  
     
